from py_glpi.connection import GLPISession, ResourceId
from dataclasses import dataclass, field
from warnings import warn

from typing import Literal, TypeVar, Generic, Type, Iterator
from typing_extensions import Self

from arrow import get
from leopards import Q

class ResourceNotFound(Exception):
    """ Raised when a resource get method couldn't find the matching item for the given id. """

class MultipleGetResult(Exception):
    """ Raised when a get operation returns more than one result (a list instance). """

class ItemCreationError(Exception):
    """ Raised if a item creation request fails. """

EXPAND_DROPDOWNS_WARNING = "Usage of expand_dropdowns migth produce unexpected behaviour for the Resource associated model.", RuntimeWarning
""" Message and type for the warning generated when expand dropdowns is used. """


@dataclass(repr=False, kw_only=True)
class GLPIItem:
    """ Basic base class for a GLPI Item, all the defined items that inherit this class should be dataclasses. This class ships the id attribute by default.
    
    IMPORTANT: The defined class name needs to match the GLPI itemtype name.
    """

    id: int
    """ Predefined id attribute, this should always exists in any GLPI Item representation. """
    
    __init_kwargs__: dict = field(default=None, init=False)
    """ All kwargs used as argument on the instanciation method. """

    connection: GLPISession = field(default=None, init=False)
    """ GLPI Connection used for generating this Item. """

    def __dates_to_arrow__(self):
        """ Attempts to transform every date attribute to arrow type. """
        for attr in self.__dict__:
            date = self.__getattribute__(attr)
            self.__setattr__(attr, get(date)) if 'date' in attr and date else None

    def post_initialization(self):
        """ This method will be executed after the GLPI Item is initialized, can be used to transform data types and format the response data, or registering new attributes. """
        ...

    def __post_init__(self):
        """ Dataclass shipped method for post initalization, shouldn't be modified. """
        assert hasattr(self, 'id'), "Every GLPI Item requires to have an ID attribute."
        self.__dates_to_arrow__()
        self.post_initialization()

    def __repr__(self) -> str:
        return f"<GLPI {self.__class__.__name__}({', '.join(f'{k}={v}' for k, v in self.__dict__.items() if not k.startswith('__') and k not in ('connection', 'resource'))})>"
    
    def as_dict(self):
        """ Used for returning a formatted dict representation of the GLPI item instance. """
        return self.__dict__

    def get_api_object(self):
        """ Returns the GLPI API item used to initialize this class """
        return self.__init_kwargs__

@dataclass(repr=False)
class FilterCriteria:
    
    field_uid: str
    operation: Literal['contains', 'equals', 'notequals', 'lessthan', 'morethan', 'under', 'notunder']
    value: str | int
    
    def __post_init__(self):
        self.link = None
        self.field_uid = self.field_uid.lower()
        self._field_uid = self.field_uid
        self.operation = self.operation.upper()
        self._evaluation: list[Self] = [self]

    def set_link(self, link: Literal['OR', 'AND', 'XOR']) -> None:
        self.link = link
    
    def __and__(self, other: Self) -> Self:
        other.set_link('AND')
        self._evaluation.append(other)
        return self
    
    def __or__(self, other: Self) -> Self:
        other.set_link('OR')
        self._evaluation.append(other)
        return self

    def __xor__(self, other: Self) -> Self:
        other.set_link('XOR')
        self._evaluation.append(other)
        return self
    
    def as_dict(self):
        obj = {
            'field': self.field_uid,
            'searchtype': self.operation,
            'value': self.value
        }
        obj |= {
            'link': self.link 
        } if self.link else {}
        return obj

    def __repr__(self):
        if len(self._evaluation) == 1:
            return f"<GLPI Filter({self._field_uid} {self.operation} {self.value})>"
        else:
            return "".join([f"{obj.link if obj.link else ''}<GLPI Filter({obj._field_uid} {obj.operation} {obj.value})>" for obj in self._evaluation])


GT = TypeVar('GT')
""" GLPI Type Generic Mashup """

class ItemList(list, Generic[GT]):
    """ List with extended methods. """

    def __getitem__(self, idx: int | slice) -> GT:
        return super().__getitem__(idx)

    def __iter__(self) -> Iterator[GT]:
        return super().__iter__()
    
    def to_representation(self):
        """ Invokes the as_dict method for each GLPI Item. """
        items: ItemList[dict] = ItemList(item.as_dict() for item in self)
        return items

    def filter(self, **kwargs):
        """ Offline filtering for an item list using leopards Q.
        Refence at https://github.com/mkalioby/leopards"""
        items: ItemList[GT] = ItemList(Q(self, **kwargs))
        return items

    def exclude(self, **kwargs):
        """ Reverse Offline filtering (usage of NOT) """
        items: ItemList[GT] = ItemList(Q(self, NOT=kwargs))
        return items

    
class Resource(Generic[GT]):
    
    """ Base class for all GLPI query-able resources, every resource should have a all, search, create, save, update and delete method. """
    
    resource_type: Type[GT] = None
    """ A valid GLPI item type (Ticket, Monitor...) """

    expand_dropdowns: bool = False
    """ Expand relations on get requests (search, get, get_multiple and all), this migth cause unexpected behaviour on the resource's associated model. """
    
    def __init__(
        self, 
        glpi_connection: GLPISession,
        subitem_of: str = None,
        parent_id: int = None,
        related_field: str = None
    ):

        """ Can be instanciated as subitem resource linked to the itemtype specified in subitem_of param."""
        self.is_subitem = bool(subitem_of)
        self.subitem_of = subitem_of
        self.related_field = related_field
        self.parent_id = parent_id

        if self.expand_dropdowns:
            warn(*EXPAND_DROPDOWNS_WARNING)

        assert self.resource_type, "Every resource must specify a resource type."

        if subitem_of:
            assert parent_id, "Parent ID needs to be specified if this resource is a subitem instance."
            assert related_field, "Related field needs to be specified if this resource is a subitem instance."

        self.resource = self.resource_type.__name__
        self.resource = self.resource if not self.is_subitem else self.subitem_of + f'/{parent_id}/{self.resource}'

        self.glpi_connection = glpi_connection
        
    def instance(self, **kwargs) -> GT:
        """ Used method for instanciating the resource type to avoid attribute errors. """
        parsed_kws = {k: v for k, v in kwargs.items() if k in self.resource_type.__annotations__}
        # Ensure id is always present in instanciation
        parsed_kws.update({'id': kwargs.get('id')})

        resource_ins = self.resource_type
        resource_ins.__init_kwargs__ = kwargs
        resource_ins.connection = self.glpi_connection
        resource_ins.resource = self

        try:
            instance = resource_ins(**parsed_kws)
        except TypeError as err:
            raise TypeError(f"Unable to instance {self.resource_type.__name__} using {kwargs.keys()} -> {err}")
        return instance

    def all(self, range_from: int = 0, range_to: int = 999_999_999, **kwargs) -> ItemList[GT]:
        """ Returns all the items available from this resource, a None response means the server responded with HTTP 204. """

        resp = self.glpi_connection.get_all_items(
            self.resource, 
            expand_dropdowns=self.expand_dropdowns,
            range=f"{range_from}-{range_to}",   # Response length range, defaults to infinite
            **kwargs
        )
        
        # In case server returns no content
        if resp.status_code == 204:
            return None

        content = resp.json()

        return ItemList(map(lambda obj: self.instance(**obj), content))
        
    def search(self, criteria: FilterCriteria = None, range_from: int = 0, range_to: int = 999_999_999, **kwargs):
        """ Use the GLPI search engine to find specific instances. """

        resp = self.glpi_connection.search_items(
            self.resource, 
            criteria, 
            range=f"{range_from}-{range_to}",   # Response length range, defaults to infinite
            expand_dropdowns=self.expand_dropdowns,
            **kwargs
        )
        content: dict = resp.json()
        results_ids = [result['2'] for result in content.get('data', [])]
        if not len(results_ids):
            raise ResourceNotFound(f"Couldn't find {self.resource_type.__name__} using {criteria} criteria.") 

        return self.get_multiple(*results_ids, **kwargs)
    
    def get(self, id: ResourceId, get_hateoas=False, **kwargs):
        
        assert id, f"Requested resource id ({id}) is invalid."

        resp = self.glpi_connection.get_item(
            self.resource, 
            id, 
            expand_dropdowns=self.expand_dropdowns, 
            get_hateoas=get_hateoas,
            **kwargs
        )
        if resp.status_code == 404:
            raise ResourceNotFound(f"Couldn't find a {self.resource_type.__name__} with id {id}")

        if isinstance(resp.json(), list):
            raise MultipleGetResult(f"Expected one result for GET operation, got a list result when querying {resp.url} instead.")

        return self.instance(**resp.json())
    
    def get_multiple(self, *ids: int, **kwargs) -> ItemList[GT]:
        """ Fetch multiple items using their IDs. """

        resp = self.glpi_connection.get_multiple_items(
            *[{"itemtype": self.resource_type.__name__, "items_id": id} for id in ids],
            expand_dropdowns=self.expand_dropdowns,
            **kwargs
        )
        content = resp.json()
        return ItemList(map(lambda obj: self.instance(**obj), content))

        
    def create(self, return_instance: bool = True, set_parents_itemtype: bool = False, **kwargs):
        """ If return instance is true, a GET request will be done to get all the details of the created object, else, only the object's id will be returned. 
        
        If set_parents_itemtype is True, when a instance of this resource is created as a subitem, the parent's itemtype will be injected as an argument -> {"itemtype": [parents_itemtype]}.
        """

        if self.is_subitem:
            # Inject relation into creation kwargs
            kwargs.update({self.related_field: self.parent_id})

            # Injects the parent's itemtype as a kwarg if needed
            if set_parents_itemtype:
                kwargs.update({'itemtype': self.resource.split('/')[0]})

        resp = self.glpi_connection.create_item(self.resource, **kwargs)
        if not resp.status_code == 201:
            raise ItemCreationError(f"Failed to create {self.resource} with {kwargs}, server returned {resp.json()} with status {resp.status_code}.")
        
        id: int = resp.json().get('id')
        return self.get(id) if return_instance else id
        
    def create_multiple(self, *args: dict):
        raise NotImplementedError("Create multiple is currently not supported.")

    def __repr__(self) -> str:
        return f"<GLPI Resource({self.resource})>"


# Inherited for adding subitem method and other resource related stuff
@dataclass(repr=False, kw_only=True)
class GLPIItem(GLPIItem):

    resource: Resource = field(default=None, init=False)
    """ The resource used for generating this item. """

    def get_subitems_resource(self, subitem_type: GLPIItem, subitem_related_field: str) -> Resource[GT]:
        """ Fabric a subitem resource class for the requested subitem_type. """

        class SubitemResource(Resource):
            """ Same as a normal resource, but type-specific to avoid attribute overwriting. """
            resource_type = subitem_type

        resource: SubitemResource[GT] = SubitemResource(self.connection, self.__class__.__name__, self.id, subitem_related_field)
        return resource

    def get_related_parent(self, parent_resource: Resource, id_field: str):
        """ Query the related parent using the parent's resource class and the related id_field to this parent. """
        assert isinstance(getattr(self, id_field), int), f"Can't make a reverse lookup of the related {parent_resource.__name__} resource if the parent attribute id is expressed as a string, try disabling expand_dropdowns."
        resource: Resource = parent_resource(self.connection)
        return resource.get(getattr(self, id_field))

    def update(self, **kwargs) -> Self:
        """ Update this instance with the specified kwargs, returns the updated GLPI Item. """
        self.connection.update_item(self.__class__.__name__, self.id, **kwargs)
        return self.resource.get(self.id)

    def delete(self) -> None:
        """ Logically Deletes the GLPI item. """
        self.connection.delete_item(self.__class__.__name__, self.id)


        



from py_glpi.models import GLPIItem, Resource, ItemList
from dataclasses import dataclass

from html import unescape

from arrow import arrow
from typing import Literal

from py_glpi.resources.documents import (
    Document_Item, 
    Documents, 
    Document, 
    Path
)

# Constant index starts from 1 for GLPI -> 1 = Muy Baja, 2 = Baja...
URGENCIES = [
    'Muy Baja',
    'Baja',
    'Media',
    'Alta',
    'Muy Alta'
]

IMPACTS = [
    'Muy Bajo',
    'Bajo',
    'Medio',
    'Alto',
    'Muy Alto'
]

PRIORITIES = [
    'Muy Baja',
    'Baja',
    'Media',
    'Alta',
    'Muy Alta',
    'Primordial'
]

STATUSES = [
    'Nuevo',
    'En Curso (Asignado)',
    'En Curso (Planificado)',
    'En Espera',
    'Resuelto',
    'Cerrado', 
]

TICKET_INVOLVED_USER_TYPES = [
    'Solicitante',
    'Encargado',
    'Observador'
]


@dataclass(repr=False)
class ITILCategory(GLPIItem):

    name: str
    itilcategories_id: str | int
    """ Name / ID of the parent category """
    level: int
    """ Hierarchy Level """

    def post_initialization(self):
        try:
            self.itilcategories_id = unescape(self.itilcategories_id)
        except:
            pass

    @property
    def parent_category(self):
        parent: ITILCategory = self.get_related_parent(Categories, 'itilcategories_id')
        return parent if self.itilcategories_id else None


class Categories(Resource[ITILCategory]):
    resource_type = ITILCategory


@dataclass(repr=False)
class RequestType(GLPIItem):

    name: str
    is_active: bool | int


class Origins(Resource[RequestType]):
    resource_type = RequestType


@dataclass(repr=False)
class SLA(GLPIItem):
    
    name: str
    type: int
    """ SLA Type, where:

    0: Time to Solve
    1: Time to Respond
    """
    number_time: int
    """ Time quantity. """
    definition_time: int
    """ Time dimension. """


class TicketSLAs(Resource[SLA]):
    resource_type = SLA


@dataclass(repr=False)
class Ticket_User(GLPIItem):

    tickets_id: int | str
    users_id: int | str
    type: int | str
    """ User's participation type/role in the ticket. """

    @property
    def related_ticket(self):
        ticket: Ticket = self.get_related_parent(Tickets, 'tickets_id')
        return ticket

    @property
    def user_participation_type(self) -> str:
        return TICKET_INVOLVED_USER_TYPES[self.type-1]

    @property
    def related_user(self):
        
        # Import occurs here to avoid circular import
        from py_glpi.resources.auth import Users, User
        
        user: User = self.get_related_parent(Users, 'users_id')
        return user


class TicketUsers(Resource[Ticket_User]):
    resource_type = Ticket_User


@dataclass(repr=False)
class Ticket(GLPIItem):
    """ Data model of a GLPI Ticket, has a set of special methods and properties for easier management. """

    name: str
    """ Title of the ticket """
    content: str
    """ Ticket raw HTML/Text content """
    itilcategories_id: int | str
    """ Ticket full categorization hierarchy """
    date_creation: arrow
    closedate: arrow
    solvedate: arrow
    status: Literal[1, 2, 3, 4, 5, 6]
    urgency: Literal[1, 2, 3, 4, 5]
    impact: Literal[1, 2, 3, 4, 5]
    priority: Literal[1, 2, 3, 4, 5, 6]
    slas_id_tto: int 
    """ SLA Time to Own Ruleset """
    slas_id_ttr: int 
    """ SLA Time to Resolve Ruleset """
    users_id_recipient: int 
    """ User who created the ticket """
    requesttypes_id: int 
    """ Ticket request origin """
    is_deleted: int

    def post_initialization(self):

        self.ticket_users: Resource[Ticket_User] = self.get_subitems_resource(Ticket_User, 'tickets_id')
        self.linked_documents: Resource[Document_Item] = self.get_subitems_resource(Document_Item, 'tickets_id')

        self.content = unescape(self.content)

    @property
    def status_string(self):
        return STATUSES[self.status-1]

    @property
    def urgency_string(self):
        return URGENCIES[self.status-1]

    @property
    def impact_string(self):
        return IMPACTS[self.status-1]

    @property
    def priority_string(self):
        return PRIORITIES[self.status-1]


    def add_applicant(self, user_id: int):
        """ Add an applicant to this ticket. """
        self.ticket_users.create(users_id=user_id, type=1)
        return True

    def add_observer(self, user_id: int):
        """ Add an observer to this ticket. """
        self.ticket_users.create(users_id=user_id, type=2)
        return True

    def add_responsible(self, user_id: int):
        """ Add a responsible to this ticket. """
        self.ticket_users.create(users_id=user_id, type=3)
        return True

    def get_involved_users(self):
        """ Returns the users involved into this ticket. """
        return self.ticket_users.all()

    def link_document(self, document_id: int):
        """ Link an existing document to this ticket. """
        self.linked_documents.create(
            set_parents_itemtype=True, documents_id=document_id, items_id=self.id
        )
        return True

    def attach_document(self, file_path: Path, file_title: str):
        """ Creates a document and links it to this ticket. """
        document_id = Documents(self.connection).create(file_path, file_title)
        self.link_document(document_id)
        return True

    def get_linked_documents(self) -> ItemList[Document]:
        return ItemList(di.related_document for di in self.linked_documents.all())

    @property
    def category(self) -> ITILCategory | None:
        """ ITIL Category related object. """
        return self.get_related_parent(Categories, 'itilcategories_id') if self.itilcategories_id else None

    @property
    def origin(self) -> RequestType | None:
        """ Request Type related object. """
        return self.get_related_parent(Origins, 'requesttypes_id') if self.requesttypes_id else None

    @property
    def tto_sla(self) -> SLA | None:
        """ TTO SLA related object. """
        return self.get_related_parent(TicketSLAs, 'slas_id_tto') if self.slas_id_tto else None

    @property
    def ttr_sla(self) -> SLA | None:
        """ TTR SLA related object. """
        return self.get_related_parent(TicketSLAs, 'slas_id_ttr') if self.slas_id_ttr else None

    def delete(self) -> None:
        return self.update(is_deleted=1)


class Tickets(Resource[Ticket]):
    resource_type = Ticket

    # Overload parameters for easier creation and assertion.
    def create(
        self,
        name: str,
        content: str,
        itilcategories_id: int,
        urgency: Literal[1, 2, 3, 4, 5] = 1,
        impact: Literal[1, 2, 3, 4, 5] = 1,
        priority: Literal[1, 2, 3, 4, 5, 6] = 1,
        slas_id_tto: int = 1,
        slas_id_ttr: int = 1,
        requesttypes_id: int = 1,
        return_instance: bool = True,
        set_parents_itemtype: bool = False,
    ):
        return super().create(
            name = name,
            content = content,
            itilcategories_id = itilcategories_id,
            urgency = urgency,
            impact = impact,
            priority = priority,
            slas_id_tto = slas_id_tto,
            slas_id_ttr = slas_id_ttr,
            requesttypes_id = requesttypes_id,
            return_instance = return_instance,
            set_parents_itemtype = set_parents_itemtype,
        )





from py_glpi.models import GLPIItem, Resource
from py_glpi.helpers import get_item_url
from dataclasses import dataclass
from arrow import arrow
from mimetypes import guess_type

from pathlib import Path
from json import dumps

class FileUploadError(Exception):
    """ Raised when the file uploading process fails. """


@dataclass(repr=False)
class Document(GLPIItem):

    name: str
    filename: str
    filepath: str
    """ GLPI internal filepath """
    mime: str
    """ Document MimeType """
    is_deleted: bool
    users_id: str | int
    """ User who created the document """
    sha1sum: str
    date_creation: arrow

    def download(self, target_path: Path, **kwargs):

        resp = self.connection.download_document(self.id, **kwargs)
        with open(target_path, 'wb+') as file:
            for chunk in resp.iter_content(chunk_size=8192):
                file.write(chunk)

        return target_path


class Documents(Resource[Document]):
    resource_type = Document

    # Override method to access required POST files attribute.

    def create(self, file: Path, file_title: str, return_instance: bool = False):
        """ Uploads a file to GLPI

        -file: The File path we want to upload.
        -file_title: The File Title that will be displayed on GLPI.
        """

        url = get_item_url(self.glpi_connection.api_url, self.resource)

        upload_manifest = {
            'input': {
                'name': file_title,
                '_filename': [file.name]
            }
        }

        file_payload = {
            'filename[0]': (file.name, open(file, 'rb'), guess_type(file)),
            'uploadManifest': (None, dumps(upload_manifest), 'application/json')
        }

        resp = self.glpi_connection.session.post(
            url,
            files=file_payload,
            data=upload_manifest
        )

        content = resp.json()

        if resp.status_code != 201:
            raise FileUploadError(f"Failed to upload {file.name}, server response was {content} with status {resp.status_code}.")

        file_id: int = content.get('id')
        return self.get(file_id) if return_instance else file_id


@dataclass(repr=False)
class Document_Item(GLPIItem):

    items_id: int
    itemtype: str
    documents_id: int

    @property
    def related_document(self) -> Document:
        """ Document related object """
        return self.get_related_parent(Documents, 'documents_id')


class DocumentRelatedItems(Resource[Document_Item]):
    resource_type = Document_Item


def get_item_url(base_url: str, item_type: str):
    return f"{base_url}/{item_type}/"
    
def parse_kwargs(kwargs: dict) -> dict:
    kwargs = {k: v for k, v in kwargs.items() if not (isinstance(v, bool) and not v)}
    return kwargs


import pytest

from tests.settings import get_connection, FIXED_DOCUMENT_ID_FOR_TESTING, BASE_DIR
from py_glpi.resources.documents import Documents, DocumentRelatedItems, GLPIItem
from py_glpi.resources.tickets import Tickets

connection = get_connection()

def test_document_all():
    document = Documents(connection).all()
    assert document


def test_document_download():
    document = Documents(connection).get(FIXED_DOCUMENT_ID_FOR_TESTING).download(BASE_DIR / 'files' / 'downloaded' / 'test.png')
    assert document


def test_document_create():
    document = Documents(connection).create(BASE_DIR / 'files' / 'fox.png', 'Firefox Icon', True)
    assert document
    

def test_document_item_parent():
    document_items: GLPIItem = DocumentRelatedItems(connection).get(FIXED_DOCUMENT_ID_FOR_TESTING)
    tickets = document_items.get_related_parent(Tickets, 'items_id')
    assert tickets
    
    
def test_document_item_all():
    document_items: GLPIItem = DocumentRelatedItems(connection).all()
    assert document_items


def test_document_item_parent():
    document_item: GLPIItem = DocumentRelatedItems(connection).get(FIXED_DOCUMENT_ID_FOR_TESTING)
    tickets = document_item.get_related_parent(Tickets, 'items_id')
    assert tickets



from py_glpi.models import GLPIItem, Resource, ItemList
from dataclasses import dataclass

from py_glpi.resources.tickets import Ticket, Ticket_User, TicketUsers

@dataclass(repr=False)
class UserEmail(GLPIItem):

    users_id: int
    is_dynamic: bool
    is_default: bool
    email: str


@dataclass(repr=False)
class User(GLPIItem):

    name: str
    firstname: str
    realname: str
    is_active: bool
    mobile: str
    auths_id: str
    """ Domain used for Active Directory authentication """
    user_dn: str
    """ User's active directory DN for debugging """

    def post_initialization(self):
        self.emails: Resource[UserEmail] = self.get_subitems_resource(UserEmail, 'users_id')
        self.tickets: TicketUsers = self.get_subitems_resource(Ticket_User, 'users_id')

    def get_observing_tickets(self) -> ItemList[Ticket]:
        """ Returns all the tickets (detailed) where this user is involved as the observer. """
        tickets = self.tickets.all().filter(type=2)
        return ItemList(item.related_ticket for item in tickets)

    def get_requested_tickets(self) -> ItemList[Ticket]:
        """ Returns all the tickets (detailed) where this user is involved as the applicant. """
        tickets = self.tickets.all().filter(type=1)
        return ItemList(item.related_ticket for item in tickets)

    def get_assigned_tickets(self) -> ItemList[Ticket]:
        """ Returns all the tickets (detailed) where this user is involved as the responsible. """
        tickets = self.tickets.all().filter(type=3)
        return ItemList(item.related_ticket for item in tickets)

    def as_dict(self):
        obj = self.__dict__
        obj['emails'] = self.emails.all().to_representation()
        return obj

class Users(Resource[User]):
    resource_type = User


    





import pytest

from tests.settings import (
    get_connection, 
    FIXED_TICKET_ID_FOR_TESTING, 
    FIXED_USER_ID_FOR_TESTING, 
    FIXED_DOCUMENT_ID_FOR_TESTING, 
    BASE_DIR
)
from py_glpi.resources.tickets import Tickets, TicketUsers, Categories, Origins, TicketSLAs
from py_glpi.models import FilterCriteria

connection = get_connection()

def test_ticket_get():
    ticket = Tickets(connection).get(FIXED_TICKET_ID_FOR_TESTING)
    assert ticket
    

def test_ticket_search():
    filter = FilterCriteria("Ticket.TicketValidation.User.name", "contains", "")
    search = Tickets(connection).search(filter)
    assert search


def test_ticket_user_search():
    filter = FilterCriteria("Ticket_User.User.id", "morethan", "1")
    search = TicketUsers(connection).search(filter)
    assert search


def test_ticket_user_parent_get():
    ticket_user = TicketUsers(connection).get(1)
    user = ticket_user.related_user
    ticket = ticket_user.related_ticket
    assert user
    assert ticket


def test_ticket_get_assigned_users():
    ticket = Tickets(connection).get(FIXED_TICKET_ID_FOR_TESTING)
    users = ticket.get_involved_users()
    assert users
    

def test_ticket_update():
    ticket = Tickets(connection).get(FIXED_TICKET_ID_FOR_TESTING)
    ticket = ticket.update(name="Test Ticket API", priority=2)
    assert ticket


def test_ticket_category_all():
    categories = Categories(connection).all()
    assert categories


def test_ticket_category_filtering():
    categories = Categories(connection).all().filter(level=3)
    assert categories


def test_ticket_origin_all():
    origins = Origins(connection).all()
    assert origins


def test_ticket_sla_all():
    origins = TicketSLAs(connection).all()
    assert origins


def test_ticket_documents():
    documents = Tickets(connection).get(FIXED_TICKET_ID_FOR_TESTING).get_linked_documents()
    print(documents)
    assert documents


def test_attach_ticket_document():
    assert Tickets(connection).get(FIXED_TICKET_ID_FOR_TESTING).attach_document(
        BASE_DIR / 'files' / 'fox.png', 
        'Firefox Icon'
    )

def test_ticket_get_parents():
    ticket = Tickets(connection).get(FIXED_TICKET_ID_FOR_TESTING)
    category = ticket.category
    origin = ticket.origin
    tto_sla = ticket.tto_sla
    ttr_sla = ticket.ttr_sla
    assert category 
    assert origin 
    assert tto_sla 
    assert ttr_sla


def test_ticket_creation_assignment_and_document_linking():
    resource = Tickets(connection)
    ticket = resource.create(
        name = "PyTest GLPI Ticket",
        content= "<p>Ticket creado por un Test del API Rest</p>",
        itilcategories_id=12
    )
    ticket.add_applicant(FIXED_USER_ID_FOR_TESTING)
    ticket.link_document(FIXED_DOCUMENT_ID_FOR_TESTING)
    ticket.delete()
    assert ticket


from py_glpi.helpers import get_item_url, parse_kwargs
from typing import TypeAlias, Literal
from requests import Session, Response

GlpiType: TypeAlias = str
ResourceId: TypeAlias = int
Url: TypeAlias = str

class AuthError(Exception):
    """ Raised if there's an error in the GLPI Session authentication process. """
        
class ItemDoesNotExists(Exception):
    """ Raised if the requested GLPI item does not exists. (ERROR_RESOURCE_NOT_FOUND_NOR_COMMONDBTM) """
        
class Unauthorized(Exception):
    """ Raised if a request produces a 401 error. """
        
class _Session(Session):
    """ Modified HTTP Session class to raise exception on forbidden response status code. """
    
    def __init__(self, **request_kwargs):
        super().__init__()
        self.request_kwargs = request_kwargs
    
    def request(self, *args, **kwargs):
        kwargs |= self.request_kwargs
        response = super().request(*args, **kwargs)
        if response.status_code == 403:
            raise Unauthorized
        return response

class URLs:

    """ Specific GLPI API URL Suffixes. """

    LOGIN: Url = 'initSession'
    RESOURCE_SEARCH: Url = 'search'
    SEARCH_OPTIONS: Url = 'listSearchOptions'
    GET_MULTIPLE_ITEMS: Url = 'getMultipleItems'
    DOCUMENTS: Url = 'Document'

class GLPISession:
    
    """ A Class that wraps a GLPI connection with all it's API endpoints, instanciating this class will create an authenticated GLPI HTTP Session using the specified authentication type."""
    
    def __init__(
        self, 
        api_url: Url = None,
        app_token: str = None, 
        auth_type: Literal["basic", "user_auth"] = "basic", 
        user: str = None, 
        password: str = None, 
        user_token: str = None,
        **session_request_kwargs
    ):
        
        # First session is produced to make the necessary login.
        self.session = _Session(**session_request_kwargs)

        assert api_url, "API URL is required for GLPI Session."
        self.api_url: Url = api_url

        assert app_token, "App token is required for GLPI Session."

        assert auth_type in ["basic", "user_token"], f"Invalid auth type {auth_type}."

        if auth_type == "basic":
            assert user and password, "A user and a password are required when using basic authentication."

        elif auth_type == "user_token":
            assert user_token, "A user token is required when using user_token authentication."

        self.auth_type = auth_type
        self.user = user
        self.__pasword = password
        self.__app_token = app_token
        self.user_token = user_token
        
        # Then, request headers are built using the necessary obtained tokens.
        self.session.headers = self.__get_request_headers()
        
    def __get_authorization(self):
        
        auth_type = self.auth_type
        if auth_type == 'user_token':
            return f"user_token {self.user_token}"
        elif auth_type == 'basic':
            from base64 import b64encode
            b64_bytes = b64encode(bytes(f"{self.user}:{self.__pasword}", encoding='utf8'))
            return f"basic {b64_bytes.decode('ascii')}"
        
    def __get_session_token(self):
        
        url = get_item_url(self.api_url, URLs.LOGIN)
        authorization_string = self.__get_authorization()
        rq = self.session.get(
            url, 
            headers = {
                'Content-Type': 'application/json',
                'App-Token': self.__app_token,
                'Authorization': authorization_string,
            }
        )
            
        if rq.status_code == 401:
            raise AuthError(f"Provided credentials are invalid. {rq.json()}")

        elif rq.status_code == 400:
            raise AuthError(f"Failed to obtain session token for GLPI Session, server response was {rq.json()} with status {rq.status_code}.")
        
        token = rq.json().get('session_token')
            
        return token
        
    def __get_request_headers(self):
        session_token = self.__get_session_token()
        return {
            "Accept": "application/json",
            "App-Token": self.__app_token,
            "Session-Token": session_token
        }
        
    def request(func):
        """ Decorator used for parsing endpoint responses. """
        
        def inner(self, *args, **kwargs) -> Response:
            try:
                call = func(self, *args, **kwargs)
                if call.status_code == 400 and call.json()[0] == 'ERROR_RESOURCE_NOT_FOUND_NOR_COMMONDBTM':
                    raise ItemDoesNotExists(
                        f'Requested item type on {func.__name__} call with ({args}, {kwargs}) to {call.url} does not exists.'
                    )
                return call
            except Unauthorized:
                self.session.headers = self.__get_request_headers()
                return func(self, *args, **kwargs)
            
        return inner
    
    @request
    def get_item(self, item_type: GlpiType, id: ResourceId, **kwargs):
        url = get_item_url(self.api_url, item_type) + f'/{id}'
        kwargs = parse_kwargs(kwargs)
        rq = self.session.get(url, params=kwargs)
        return rq

    @request
    def get_multiple_items(self, *item_keys: dict, **kwargs):
        """ Fetch multiple items using a key composition of {"itemtype":..., "items_id":...}. """
        url = get_item_url(self.api_url, URLs.GET_MULTIPLE_ITEMS) + '?'
        url += "&".join([f"items[{i}][itemtype]={item['itemtype']}&items[{i}][items_id]={item['items_id']}" for i, item in enumerate(item_keys)])
        kwargs = parse_kwargs(kwargs)
        rq = self.session.get(url, params=kwargs)
        return rq
    
    @request
    def get_all_items(self, item_type: GlpiType, **kwargs):
        url = get_item_url(self.api_url, item_type)
        kwargs = parse_kwargs(kwargs)
        rq = self.session.get(url, params=kwargs)
        return rq
    
    @request
    def create_item(self, item_type: GlpiType, **kwargs):
        url = get_item_url(self.api_url, item_type)
        body = {"input": kwargs}
        rq = self.session.post(url, json=body)
        return rq

    @request
    def update_item(self, item_type: GlpiType, id: ResourceId, **kwargs):
        url = get_item_url(self.api_url, item_type) + f'/{id}'
        body = {"input": kwargs}
        rq = self.session.put(url, json=body)
        return rq

    @request
    def delete_item(self, item_type: GlpiType, id: ResourceId):
        url = get_item_url(self.api_url, item_type) + f'/{id}'
        rq = self.session.delete(url)
        return rq
    
    @request
    def create_items(self, item_type: GlpiType, *args: dict):
        url = get_item_url(self.api_url, item_type)
        body = {"input": list(args)}
        rq = self.session.post(url, data=body)
        return rq
    
    @request
    def get_item_search_options(self, item_type: GlpiType):
        url = get_item_url(self.api_url, URLs.SEARCH_OPTIONS) + f'/{item_type}'
        return self.session.get(url)
    
    @request
    def search_items(self, item_type: GlpiType, criteria, **kwargs):

        search_opts: dict = self.get_item_search_options(item_type).json()
        criterias: list[dict] = [ev.as_dict() for ev in criteria._evaluation]

        # Find the specified field's id from it's name.
        for eval in criterias:
            id = [k for k, v in search_opts.items() if isinstance(v, dict) and v.get('uid', '').lower() == eval['field']]
            if not len(id):
                valid_opts = {v.get('uid'): v.get('name') for k, v in search_opts.items() if k != 'common' and isinstance(v, dict)}
                raise ValueError(f"Couldn't find a valid filtering field with uid '{eval['field']}' for itemtype '{item_type}'. Valid filtering field reference shown below:\n\n{valid_opts}")
            eval['field'] = id[0]

        url = get_item_url(self.api_url, URLs.RESOURCE_SEARCH) + item_type + '/?'
        criteria_kwargs = [[f"criteria[{i}][{k}]={v}" for k, v in criteria.items()] for i, criteria in enumerate(criterias)]
        if len(criteria_kwargs) > 0:
            kwargs = parse_kwargs(kwargs)
            url += "&".join(criteria_kwargs[0] + [f"{k}={v}" for k, v in kwargs.items()])

        url += "&forcedisplay[0]=2"
        
        return self.session.get(url)

    @request
    def download_document(self, document_id: int, **kwargs):
        url = get_item_url(self.api_url, URLs.DOCUMENTS) + f'/{document_id}?alt=media'
        return self.session.get(url, params=kwargs)
            
            
        
            
        

    
    
        